#include "..\fx\tricks.h"
#include "xreader.h"
#include <math.h>
#include <stdio.h>

byte buf[64000];
byte envmap[65536];
byte pal[256][3];

void rotate( obj *O, float x, float y, float z, float zdist, float perspective  );
void Poly( int F );
void red();
int comp( const void *a, const void *b );

void load( char *file, byte *data, unsigned int length );
void loadpal( char *file );

obj O;

struct
{
   int s, e;
   float scx, scy;
   float ecx, ecy;
} xt[200];

void main()
{
   readobj( &O, "blob.x" );

   GrabPallette();
   SetMCGA();

   // init phongmap

   load( "wormhole.raw", envmap, 65536 );
   loadpal( "wormhole.pal" );  

   for( int i=0; i<256; i++ )
      Pal(i, pal[i][0], pal[i][1], pal[i][2] );

   int *df = new int[ O.faces ];  // faces to be drawn..

   while( !kbhit() )
   {
      rotate( &O, 0.015, 0.025, 0.01, 1.5, 600 );

      int count = 0;                  // number of faces..

      for( int i=0; i<O.faces; i++ ) {
         if( O.f[i].zn > 0 )
            df[count++ ] = i;
      }

      qsort( df, count, sizeof(int), comp );

      for( i=0; i<count; i++)
         Poly( df[i] );

      int *v= (int*) vga;
      int *b= (int*) buf;

      for( i=0; i<16000; i++ ) {
         *v++ = *b;
         *b++ = 0;
      }

      WaitRetrace();
   }

   SetText();
   RestorePallette();

   delete[] df;
   delete[] O.v;
   delete[] O.f;

}

int comp( const void *a, const void *b )
{
   float za = O.f[*(int*) a].z;
   float zb = O.f[*(int*) b].z;

   if( za>zb )       
      return -1;     
   else if( za<zb )
      return 1;
   else
      return 0;
}


void rotate( obj *O, float x, float y, float z, float zd, float p )
{
   // precalc values..

   float sx = sin(x);
   float sy = sin(y);
   float sz = sin(z);

   float cx = cos(x);
   float cy = cos(y);
   float cz = cos(z);

   float mx1 = sx*sy*sz + cz*cx;
   float my1 = sx*cy;
   float mz1 = sz*cx - sx*sy*cz;

   float mx2 = cx*sy*sz - cz*sx;
   float my2 = cx*cy;
   float mz2 = -cz*sy*cx - sz*sx;

   float mx3 = -sz*cy;
   float my3 = sy;
   float mz3 = cz*cy;

   for( int i=0; i<O->vertices; i++ )
   {
      // rotate coords

      x = O->v[i].x;
      y = O->v[i].y;
      z = O->v[i].z;

      O->v[i].x = x*mx1 + y*my1 + z*mz1;
      O->v[i].y = x*mx2 + y*my2 + z*mz2;
      O->v[i].z = x*mx3 + y*my3 + z*mz3;

      // project

      z += zd;

      O->v[i].px = (x/z)*p + 160;
      O->v[i].py = (y/z)*p*.8 + 100;

      // rotate normals

      x = O->v[i].xn;
      y = O->v[i].yn;
      z = O->v[i].zn;

      O->v[i].xn = x*mx1 + y*my1 + z*mz1;
      O->v[i].yn = x*mx2 + y*my2 + z*mz2;
      O->v[i].zn = x*mx3 + y*my3 + z*mz3;
   }
   for( i=0; i<O->faces; i++) {
/*
      x = O->f[i].xn;
      y = O->f[i].yn;
      z = O->f[i].zn;

      O->f[i].xn = x*mx1 + y*my1 + z*mz1;
      O->f[i].yn = x*mx2 + y*my2 + z*mz2;
      O->f[i].zn = x*mx3 + y*my3 + z*mz3;
*/
      float p1x = O->v[O->f[i].p1].x;
      float p1y = O->v[O->f[i].p1].y;
      float p1z = O->v[O->f[i].p1].z;

      float p2x = O->v[O->f[i].p2].x;
      float p2y = O->v[O->f[i].p2].y;
      float p2z = O->v[O->f[i].p2].z;

      float p3x = O->v[O->f[i].p3].x;
      float p3y = O->v[O->f[i].p3].y;
      float p3z = O->v[O->f[i].p3].z;

      O->f[i].xn = ((p3y - p2y) * (p1z - p2z))
                 - ((p1y - p2y) * (p3z - p2z));

      O->f[i].yn = ((p3z - p2z) * (p1x - p2x))
                 - ((p1z - p2z) * (p3x - p2x));

      O->f[i].zn = ((p3y - p2y) * (p1x - p2x))
                 - ((p1y - p2y) * (p3x - p2x));

      O->f[i].z  = ( O->v[ O->f[i].p1 ].z +
                     O->v[ O->f[i].p2 ].z +
                     O->v[ O->f[i].p3 ].z ) / 3;
   }
}

void red()  
{
    for( int color = 1; color < 64; color++ )
        Pal( color, 0,0,color/2);

    for( color = 64; color < 128; color++ )
        Pal( color, 0,0,color/2);

    for( color = 128 ; color < 192; color++ )
        Pal( color, color-128,0,63);

    for( color = 192; color < 256; color++ )
        Pal( color, 63, color-192, 63);
}


void polyfix( int p1, int p2 )
{
   int y1 = O.v[p1].py;
   int y2 = O.v[p2].py;

   if( y2 < y1 )
   {
      swap( y1,y2 );
      swap( p1,p2 );
   }

   int x1  = O.v[p1].px;
   int x2  = O.v[p2].px;
   
   int scx = O.v[p1].xn *130 + 127;
   int scy = O.v[p1].yn *130 + 127;

   int ecx = O.v[p2].xn *130 + 127;
   int ecy = O.v[p2].yn *130 + 127;

   if( y2 == y1 )
   {
      xt[y1].s  = x1;
      xt[y1].e  = x2;

      xt[y1].scx = scx;
      xt[y1].scy = scy;

      xt[y1].ecx = ecx;
      xt[y1].ecy = ecy;
   }
   else
   {
      int dy    = y2-y1;

      int Step  = ((x2-x1) << 16 ) / dy;

      int xstep = ((ecx-scx) << 16 ) / dy;
      int ystep = ((ecy-scy) << 16 ) / dy;

      int X     = (x1<<16);

      int xc    = (scx<<16);
      int yc    = (scy<<16);

      y1++;
      X += Step;
      xc += xstep;
      yc += ystep;

      for( int Y = y1; Y <= y2; Y++)
      {
         if(( Y>=0 )&&( Y<200 ))
         {
            if( xt[Y].s == -1)
            {
               xt[Y].s   = (X >> 16);
               xt[Y].scx = (xc >> 16);
               xt[Y].scy = (yc >> 16);
            }
            else
            {
               xt[Y].e   = (X >> 16);
               xt[Y].ecx = (xc >> 16);
               xt[Y].ecy = (yc >> 16);
            }
         }

         X += Step;
         xc += xstep;
         yc += ystep;
      }
   }
}

void Poly( int F )
{
   for( int i=0; i<200; i++)
      xt[i].s = xt[i].e = -1;

   int p1 = O.f[F].p1;
   int p2 = O.f[F].p2;
   int p3 = O.f[F].p3;

   polyfix( p1,p2 );
   polyfix( p2,p3 );
   polyfix( p3,p1 );

   for( i=0; i<200; i++ )
   {
      int s = xt[i].s;

      if( s != -1 )
      {         
         int scx = xt[i].scx;
         int scy = xt[i].scy;
   
         int ecx = xt[i].ecx;
         int ecy = xt[i].ecy;

         int e = xt[i].e;
   
         if( e == -1)
         {
            e = s+1;
            ecx = scx;
            ecy = scy;
         }
         else if( s > e )
         {
            swap( s, e );
            swap( scx, ecx );
            swap( scy, ecy );
         }
         else if (e==s) e++;

         int dx = e-s;

         int x = (scx << 16);
         int y = (scy << 16);

         int sx = ((ecx << 16)-x) / dx;
         int sy = ((ecy << 16)-y) / dx;

         byte *b = buf + (i<<8) + (i<<6) + s;

         for( ; s<e; s++)
         {
            *b++ = envmap[ ((y>>8)&0xff00) + (x>>16) ];

            x += sx;
            y += sy;
         }
      }
   }
}

void load( char *file, byte *data, unsigned int length )
{
   FILE *fp = fopen( file, "rb" );
   fread( data, 1, length, fp);
   fclose( fp );
}

void loadpal( char *file )
{
   FILE *fp = fopen( file, "r" );

   for( int i=0; i<256; i++)
   {
      int r,g,b;

      fscanf( fp, "%d %d %d", &r,&g,&b);

      pal[i][0] = r>>2;
      pal[i][1] = g>>2;
      pal[i][2] = b>>2;
   }

   fclose( fp );
}


